# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Classifier Configuration."""


CLASSIFIER_MAXIMUM_VOCABULARY_SIZE = 60000
CLASSIFIER_MINIMUM_WORD_FREQUENCY = 2
CLASSIFIER_VALIDATION_DATA_FRACTION = 0.1
CLASSIFIER_LANGUAGE_MODEL_CYCLE_LENGTH = 15
CLASSIFIER_CLASSIFIER_CYCLE_LENGTH = 14
CLASSIFIER_LANGUAGE_MODEL_BATCH_SIZE = 32
CLASSIFIER_CLASSIFIER_BATCH_SIZE = 10
CLASSIFIER_CUDA_DEVICE_ID = -1  # set to 0 to use a GPU

CLASSIFIER_DATA_PATH = 'data'
CLASSIFIER_LANGUAGE_MODEL_PATH = 'models/language_model'
CLASSIFIER_CLASSIFIER_MODEL_PATH = 'models/classifier_model'
CLASSIFIER_LANGUAGE_MODEL_DATA_PATH = 'data/language_model_data'
CLASSIFIER_CLASSIFIER_DATA_PATH = 'data/classifier_data'
CLASSIFIER_DATAFRAME_PATH = 'data/inspire_data.df'
CLASSIFIER_PRETRAINED_LANGUAGE_MODEL_PATH = 'models/language_model/wikitext_103/fwd_wt103.h5'
CLASSIFIER_FINETUNED_LANGUAGE_MODEL_ENCODER_PATH = 'models/language_model/finetuned_language_model_encoder.h5'
CLASSIFIER_TRAINED_CLASSIFIER_PATH = 'models/classifier_model/trained_classifier_model.h5'
CLASSIFIER_WIKITEXT103_ITOS_PATH = 'models/language_model/wikitext_103/itos_wt103.pkl'
CLASSIFIER_DATA_ITOS_PATH = 'data/inspire_data_itos.pkl'

CLASSIFIER_WIKITEXT103_LANGUAGE_MODEL_URL = 'http://files.fast.ai/models/wt103/fwd_wt103.h5'
CLASSIFIER_WIKITEXT103_ITOS_URL = 'http://files.fast.ai/models/wt103/itos_wt103.pkl'



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

import datetime

from flask import Flask, jsonify, Response
from flask_apispec import use_kwargs, marshal_with, FlaskApiSpec
from inspire_classifier.api import predict_coreness
from marshmallow import fields

from . import serializers


class JsonResponse(Response):
    """"By creaitng this Response class, we force the response to always be in json, getting rid of the jsonify function."""

    @classmethod
    def force_type(cls, rv, environ=None):
        if isinstance(rv, dict):
            rv = jsonify(rv)
        return super(JsonResponse, cls).force_type(rv, environ)


def create_app():
    Flask.response_class = JsonResponse
    app = Flask(__name__)
    app.config['CLASSIFIER_BASE_PATH'] = app.instance_path
    app.config.from_object('inspire_classifier.config')
    docs = FlaskApiSpec(app)

    @app.route("/api/health")
    def date():
        """Basic endpoint that returns the date, used to check if everything is up and working."""
        now = datetime.datetime.now()
        return jsonify(now)

    docs.register(date)

    @app.route("/api/predict/coreness", methods=["POST"])
    @use_kwargs({'title': fields.Str(required=True), 'abstract': fields.Str(required=True)})
    @marshal_with(serializers.ClassifierOutputSerializer)
    def core_classifier(**kwargs):
        """Endpoint for the CORE classifier."""

        return predict_coreness(kwargs['title'], kwargs['abstract'])

    docs.register(core_classifier)

    return app

    @app.errorhandler(404)
    def page_not_found(e):
        return {
            "errors": [
                str(e)
            ]
        }, 404


app = create_app()

if __name__ == '__main__':
    app.run(host='0.0.0.0')



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""INSPIRE module aimed at automatically classifying the new papers that are added to INSPIRE, such as if they are core or not, or the arXiv category corresponding to each of them."""



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""
Get the arxiv ids of all the Core and Non-Core records in INSPIRE.
Please run the code in this snippet from within the inspirehep shell.
"""

from invenio_search import current_search_client as es
from elasticsearch.helpers import scan
import numpy as np

core = []
non_core = []

for hit in scan(es, query={"query": {"exists": {"field": "arxiv_eprints"}}, "_source": ["core", "arxiv_eprints"]},
                index='records-hep', doc_type='hep'):
    source = hit['_source']
    arxiv_eprint = source['arxiv_eprints'][0]['value']
    if source.get('core') == True:
        core.append(arxiv_eprint)
    else:
        non_core.append(arxiv_eprint)

with open('inspire_core_list.txt', 'w') as fd:
    fd.writelines("{}\n".format(arxiv_id) for arxiv_id in core)
with open('inspire_noncore_list.txt', 'w') as fd:
    fd.writelines("{}\n".format(arxiv_id) for arxiv_id in non_core)


# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.
#
# Modified from the fastai library (https://github.com/fastai/fastai).

from fastai.text import (
    num_cpus,
    Tokenizer
)
from flask import current_app
from pathlib import Path
from concurrent.futures import ProcessPoolExecutor
import re
from spacy.lang.en import English
from spacy.symbols import ORTH


def path_for(name):
    base_path = Path(current_app.config.get('CLASSIFIER_BASE_PATH') or current_app.instance_path)
    config_key = f'CLASSIFIER_{name}_PATH'.upper()

    return base_path / current_app.config[config_key]


class FastLoadTokenizer(Tokenizer):
    """
    Tokenizer which avoids redundant loading of spacy language model

    The FastAI Tokenizer class loads all the pipeline components of the spacy model which significantly increases
    loading time, especially when doing inference on CPU. This class inherits from the FastAI Tokenizer and is
    refactored to avoid redundant loading of the classifier.
    """
    def __init__(self):
        self.re_br = re.compile(r'<\s*br\s*/?>', re.IGNORECASE)
        self.tok = English()
        for w in ('<eos>', '<bos>', '<unk>'):
            self.tok.tokenizer.add_special_case(w, [{ORTH: w}])

    def proc_all(self, ss):
        return [self.proc_text(s) for s in ss]

    def proc_all_mp(self, ss, ncpus=None):
        ncpus = ncpus or num_cpus() // 2
        with ProcessPoolExecutor(ncpus) as executor:
            return sum(executor.map(self.proc_all, ss), [])



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""
Get Core and Non-Core records starting from an earliest date from INSPIRE.
Please run the code in this snippet from within the inspirehep shell.
"""

import datetime
from invenio_db import db
from invenio_records.models import RecordMetadata
import json
from sqlalchemy import (
    and_,
    cast,
    not_,
    or_,
    type_coerce
)
from sqlalchemy.dialects.postgresql import JSONB


STARTING_DATE = datetime.datetime(2016, 1, 1, 0, 0, 0)

base_query = db.session.query(RecordMetadata).with_entities(RecordMetadata.json['titles'][0]['title'], RecordMetadata.json['abstracts'][0]['value'])
filter_by_date = RecordMetadata.created >= STARTING_DATE
has_title_and_abstract = and_(type_coerce(RecordMetadata.json, JSONB).has_key('titles'), type_coerce(RecordMetadata.json, JSONB).has_key('abstracts'))
filter_deleted_records = or_(not_(type_coerce(RecordMetadata.json, JSONB).has_key('deleted')), not_(RecordMetadata.json['deleted'] == cast(True, JSONB)))
only_literature_collection = type_coerce(RecordMetadata.json, JSONB)['_collections'].contains(['Literature'])

only_core_records = type_coerce(RecordMetadata.json, JSONB)['core'] == cast(True, JSONB)
only_noncore_records = or_(type_coerce(RecordMetadata.json, JSONB)['core'] == cast(False, JSONB), not_(type_coerce(RecordMetadata.json, JSONB).has_key('core')))

core_query_results = base_query.filter(filter_by_date, only_core_records, has_title_and_abstract, filter_deleted_records, only_literature_collection)
noncore_query_results = base_query.filter(filter_by_date, only_noncore_records, has_title_and_abstract, filter_deleted_records, only_literature_collection)

with open('inspire_core_records.jsonl', 'w') as fd:
    for title, abstract in core_query_results:
        fd.write(json.dumps({
            'title': title,
            'abstract': abstract,
        }) + '\n')
with open('inspire_noncore_records.jsonl', 'w') as fd:
    for title, abstract in noncore_query_results:
        fd.write(json.dumps({
            'title': title,
            'abstract': abstract,
        }) + '\n')



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.
#
# Modified from the fastai library (https://github.com/fastai/fastai).

"""Classifier Domain Preprocessors."""


import collections
from fastai.text import partition_by_cores
import html
from inspire_classifier.utils import FastLoadTokenizer
import numpy as np
import pandas as pd
import pickle
import re
import sklearn


BOS = 'xbos'  # beginning-of-sentence tag
FLD = 'xfld'  # data field tag
re1 = re.compile(r'  +')


def split_and_save_data_for_language_model_and_classifier(dataframe_path, language_model_data_dir, classifier_data_dir,
                                                          val_fraction=0.1):
    """
    Args:
        dataframe_path: The path to the pandas dataframe containing the records. The dataframe should have one
                        column containing the title and abstract text appended (title + abstract). The second
                        column should contain the label as an integer (0: Rejected, 1: Non-Core, 2: Core).
        language_model_data_dir: Directory to store language model data.
        classifier_data_dir: Directory to save classifier data.
        val_fraction: the fraction of data to use as the validation set.
    """
    inspire_data = pd.read_pickle(dataframe_path)

    # Shuffle the data
    inspire_data = inspire_data.sample(frac=1).reset_index(drop=True)
    # Swap the columns so that the labels are Column 0 and the text is Column 1 (and remove any additional columns)
    inspire_data = inspire_data[['labels', 'text']]

    training_dataframe, validation_dataframe = sklearn.model_selection.train_test_split(
        inspire_data, test_size=val_fraction)

    training_dataframe = training_dataframe.reset_index(drop=True)
    validation_dataframe = validation_dataframe.reset_index(drop=True)

    # Save the data for the classifier
    training_dataframe.to_csv(classifier_data_dir / 'training_data.csv', header=False, index=False)
    validation_dataframe.to_csv(classifier_data_dir / 'validation_data.csv', header=False, index=False)

    training_texts = np.array(training_dataframe['text'])
    validation_texts = np.array(validation_dataframe['text'])

    column_names = ['labels', 'text']
    training_dataframe_for_language_model = pd.DataFrame({'text': training_texts, 'labels': [0] * len(training_texts)},
                                                         columns=column_names)
    validation_dataframe_for_language_model = pd.DataFrame({'text': validation_texts,
                                                            'labels': [0] * len(validation_texts)},
                                                           columns=column_names)

    training_dataframe_for_language_model.to_csv(language_model_data_dir / 'training_data.csv', header=False,
                                                 index=False)
    validation_dataframe_for_language_model.to_csv(language_model_data_dir / 'validation_data.csv', header=False,
                                                   index=False)


def generate_and_save_language_model_tokens(language_model_data_dir):
    training_dataframe = pd.read_csv(language_model_data_dir / 'training_data.csv', header=None)
    validation_dataframe = pd.read_csv(language_model_data_dir / 'validation_data.csv', header=None)

    training_tokens, training_labels = get_texts(training_dataframe)
    validation_tokens, validation_labels = get_texts(validation_dataframe)

    assert len(training_tokens) == len(training_dataframe)

    np.save(language_model_data_dir / 'training_tokens.npy', training_tokens)
    np.save(language_model_data_dir / 'validation_tokens.npy', validation_tokens)
    np.save(language_model_data_dir / 'training_labels.npy', training_labels)
    np.save(language_model_data_dir / 'validation_labels.npy', validation_labels)


def map_and_save_tokens_to_ids_for_language_model(language_model_data_dir, data_itos_path, max_vocab_size=60000,
                                                  minimum_frequency=2):
    """
    Args:
        language_model_data_dir: Directory for language model data.
        data_itos_path: The path to save the data ITOS which maps the words in the vocabulary to numerical indices.
        max_vocab_size: The maximum size of the vocabulary (default: 60000).
        minimum_frequency: The minimum frequency that a word has to occur to be included in the vocabulary. This
                            prevents including words which occur rarely (default: 2)
    """
    training_tokens = np.load(language_model_data_dir / 'training_tokens.npy')
    validation_tokens = np.load(language_model_data_dir / 'validation_tokens.npy')

    word_frequency = collections.Counter(p for o in training_tokens for p in o)
    inspire_data_itos = [o for o, c in word_frequency.most_common(max_vocab_size) if c > minimum_frequency]
    inspire_data_itos.insert(0, '_pad_')
    inspire_data_itos.insert(0, '_unk_')
    inspire_data_stoi = collections.defaultdict(lambda: 0, {v: k for k, v in enumerate(inspire_data_itos)})

    training_token_ids = np.array([[inspire_data_stoi[o] for o in p] for p in training_tokens])
    validation_token_ids = np.array([[inspire_data_stoi[o] for o in p] for p in validation_tokens])

    np.save(language_model_data_dir / 'training_token_ids.npy', training_token_ids)
    np.save(language_model_data_dir / 'validation_token_ids.npy', validation_token_ids)
    pickle.dump(inspire_data_itos, open(data_itos_path, 'wb'))


def generate_and_save_classifier_tokens(classifier_data_dir):
    training_dataframe = pd.read_csv(classifier_data_dir / 'training_data.csv', header=None)
    validation_dataframe = pd.read_csv(classifier_data_dir / 'validation_data.csv', header=None)

    training_tokens, training_labels = get_texts(training_dataframe)
    validation_tokens, validation_labels = get_texts(validation_dataframe)

    assert len(training_tokens) == len(training_dataframe)

    np.save(classifier_data_dir / 'training_tokens.npy', training_tokens)
    np.save(classifier_data_dir / 'validation_tokens.npy', validation_tokens)
    np.save(classifier_data_dir / 'training_labels.npy', training_labels)
    np.save(classifier_data_dir / 'validation_labels.npy', validation_labels)


def map_and_save_tokens_to_ids_for_classifier(classifier_data_dir, data_itos_path):
    training_tokens = np.load(classifier_data_dir / 'training_tokens.npy')
    validation_tokens = np.load(classifier_data_dir / 'validation_tokens.npy')

    inspire_data_itos = pickle.load(open(data_itos_path, 'rb'))
    inspire_data_stoi = collections.defaultdict(lambda: 0, {v: k for k, v in enumerate(inspire_data_itos)})

    training_token_ids = np.array([[inspire_data_stoi[o] for o in p] for p in training_tokens])
    validation_token_ids = np.array([[inspire_data_stoi[o] for o in p] for p in validation_tokens])

    np.save(classifier_data_dir / 'training_token_ids.npy', training_token_ids)
    np.save(classifier_data_dir / 'validation_token_ids.npy', validation_token_ids)


def get_texts(df):
    labels = df[0].values.astype(np.int64)
    texts = f'\n{BOS} {FLD} 1 ' + df[1].astype(str)
    texts = list(texts.apply(fixup).values)

    tokens = FastLoadTokenizer().proc_all_mp(partition_by_cores(texts))
    return tokens, list(labels)


def fixup(x):
    x = x.replace('#39;', "'").replace('amp;', '&').replace('#146;', "'").replace(
        'nbsp;', ' ').replace('#36;', '$').replace('\\n', "\n").replace('quot;', "'").replace(
        '<br />', "\n").replace('\\"', '"').replace('<unk>', 'u_n').replace(' @.@ ', '.').replace(
        ' @-@ ', '-').replace('\\', ' \\ ')
    return re1.sub(' ', html.unescape(x))



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

import json
import numpy as np
import pandas as pd

inspire_core_list_path = 'inspire_core_list.txt'
inspire_noncore_list_path = 'inspire_noncore_list.txt'
inspire_harvested_data_path = 'inspire_harvested_data.jsonl'
inspire_core_data_path = 'inspire_core_records.jsonl'
inspire_noncore_data_path = 'inspire_noncore_records.jsonl'
save_path = 'inspire_data.df'

with open(inspire_core_list_path, 'r') as fd:
    inspire_core_arxiv_ids = set(arxiv_id.strip() for arxiv_id in fd.readlines())
with open(inspire_noncore_list_path, 'r') as fd:
    inspire_noncore_arxiv_ids = set(arxiv_id.strip() for arxiv_id in fd.readlines())

def rejected_data(harvested_data_path):
    with open(harvested_data_path, 'r') as fd:
        for line in fd:
            try:
                record = json.loads(line)
                if not (record['arxiv_identifier'] in inspire_core_arxiv_ids) and \
                        not (record['arxiv_identifier'] in inspire_noncore_arxiv_ids):
                    del record['arxiv_identifier']
                    yield record
            except:
                continue

def core_data():
    with open(inspire_core_data_path, 'r') as fd:
        for line in fd:
            yield json.loads(line)

def noncore_data():
    with open(inspire_noncore_data_path, 'r') as fd:
        for line in fd:
            yield json.loads(line)

rejected_df = pd.DataFrame(rejected_data(inspire_harvested_data_path))
rejected_df['labels'] = 0
noncore_df = pd.DataFrame(core_data())
noncore_df['labels'] = 1
core_df = pd.DataFrame(noncore_data())
core_df['labels'] = 2

inspire_data = pd.concat([rejected_df, noncore_df, core_df], ignore_index=True)
#XXX: The below data concatenation method assumes the current title and abstrat
#     combination approach i.e. using a simple whitespace. However, it's better
#     to insert a special token such as ' <endTitle> ' instead of the single
#     whitespace. This would require training our models again, as well as making
#     changes in the predict_coreness API to make the text consistent with this.
inspire_data['text'] = inspire_data['title'] + ' ' + inspire_data['abstract']
inspire_data.to_pickle(save_path)



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.
#
# Modified from the fastai library (https://github.com/fastai/fastai).

"""Classifier Domain Models."""

import collections
import torch
import torch.optim as optim
from fastai.text import (
    accuracy,
    DataLoader,
    get_rnn_classifer,
    LanguageModelLoader,
    LanguageModelData,
    load_model,
    ModelData,
    RNN_Learner,
    T,
    TextDataset,
    TextModel,
    to_gpu,
    to_np,
    save_model,
    seq2seq_reg,
    SortishSampler,
    SortSampler,
    Variable
)
from functools import partial
from inspire_classifier.utils import FastLoadTokenizer
import numpy as np
import pickle


class LanguageModel(object):
    def __init__(self, training_data_ids_path, validation_data_ids_path, language_model_model_dir,
                 data_itos_path, cuda_device_id=0, batch_size=32, dropout_multiplier=0.7):
        torch.cuda.set_device(cuda_device_id)
        self.use_cuda = True if cuda_device_id >= 0 else False

        self.inspire_data_itos = pickle.load(open(data_itos_path, 'rb'))
        self.vocabulary_size = len(self.inspire_data_itos)

        number_of_backpropagation_through_time_steps = 70
        number_of_hidden_units = 1150
        number_of_layers = 3
        self.embedding_size = 400
        optimization_function = partial(optim.Adam, betas=(0.8, 0.99))

        training_token_ids = np.load(training_data_ids_path)
        training_token_ids = np.concatenate(training_token_ids)
        validation_token_ids = np.load(validation_data_ids_path)
        validation_token_ids = np.concatenate(validation_token_ids)

        training_dataloader = LanguageModelLoader(nums=training_token_ids, bs=batch_size,
                                                  bptt=number_of_backpropagation_through_time_steps)
        validation_dataloader = LanguageModelLoader(nums=validation_token_ids, bs=batch_size,
                                                    bptt=number_of_backpropagation_through_time_steps)
        model = LanguageModelData(path=language_model_model_dir, pad_idx=1, n_tok=self.vocabulary_size,
                                  trn_dl=training_dataloader, val_dl=validation_dataloader, bs=batch_size,
                                  bptt=number_of_backpropagation_through_time_steps)

        dropouts = np.array([0.25, 0.1, 0.2, 0.02, 0.15]) * dropout_multiplier

        self.learner = model.get_model(opt_fn=optimization_function, emb_sz=self.embedding_size,
                                       n_hid=number_of_hidden_units, n_layers=number_of_layers, dropouti=dropouts[0],
                                       dropout=dropouts[1], wdrop=dropouts[2], dropoute=dropouts[3],
                                       dropouth=dropouts[4])
        self.learner.reg_fn = partial(seq2seq_reg, alpha=2, beta=1)
        self.learner.clip = 0.3
        self.learner.metrics = [accuracy]

    def load_pretrained_language_model_weights(self, pretrained_language_model_path, wikitext103_itos_path):
        weights = torch.load(pretrained_language_model_path, map_location=lambda storage, loc: storage)

        encoder_weights = to_np(weights['0.encoder.weight'])
        row_m = encoder_weights.mean(0)

        wikitext103_itos = pickle.load(open(wikitext103_itos_path, 'rb'))
        wikitext103_stoi = collections.defaultdict(lambda: -1, {v: k for k, v in enumerate(wikitext103_itos)})

        nw = np.zeros((self.vocabulary_size, self.embedding_size), dtype=np.float32)
        for i, w in enumerate(self.inspire_data_itos):
            r = wikitext103_stoi[w]
            if r >= 0:
                nw[i] = encoder_weights[r]
            else:
                nw[i] = row_m

        weights['0.encoder.weight'] = T(nw, cuda=self.use_cuda)
        weights['0.encoder_with_dropout.embed.weight'] = T(np.copy(nw), cuda=self.use_cuda)
        weights['1.decoder.weight'] = T(np.copy(nw), cuda=self.use_cuda)

        self.learner.model.load_state_dict(weights)

    def train(self, finetuned_language_model_encoder_save_path, learning_rate=1e-3, weight_decay=1e-7, cycle_length=15):
        self.learner.freeze_to(-1)
        self.learner.fit(learning_rate / 2, n_cycle=1, wds=weight_decay, use_clr=(32, 2), cycle_len=1)
        self.learner.unfreeze()
        self.learner.fit(learning_rate, n_cycle=1, wds=weight_decay, use_clr=(20, 10), cycle_len=cycle_length)
        save_model(self.learner.model[0], finetuned_language_model_encoder_save_path)


class Classifier(object):
    def __init__(self, data_itos_path, cuda_device_id=0, dropout_multiplier=0.5, number_of_classes=3):
        torch.cuda.set_device(cuda_device_id)

        inspire_data_itos = pickle.load(open(data_itos_path, 'rb'))
        self.vocabulary_size = len(inspire_data_itos)
        self.inspire_data_stoi = collections.defaultdict(
            lambda: 0, {str(v): int(k) for k, v in enumerate(inspire_data_itos)})

        dropouts = np.array([0.4, 0.5, 0.05, 0.3, 0.4]) * dropout_multiplier

        number_of_back_propagation_through_time_steps = 70
        number_of_hidden_units = 1150
        number_of_layers = 3
        embedding_size = 400

        self.model = get_rnn_classifer(bptt=number_of_back_propagation_through_time_steps,
                                       max_seq=20 * number_of_back_propagation_through_time_steps,
                                       n_class=number_of_classes, n_tok=self.vocabulary_size, emb_sz=embedding_size,
                                       n_hid=number_of_hidden_units, n_layers=number_of_layers, pad_token=1,
                                       layers=[embedding_size * 3, 50, number_of_classes], drops=[dropouts[4], 0.1],
                                       dropouti=dropouts[0], wdrop=dropouts[1], dropoute=dropouts[2],
                                       dropouth=dropouts[3])

        self.tokenizer = FastLoadTokenizer()

    def load_training_and_validation_data(self, training_data_ids_path, training_data_labels_path,
                                          validation_data_ids_path, validation_data_labels_path, classifier_data_dir,
                                          batch_size=10):
        training_token_ids = np.load(training_data_ids_path)
        validation_token_ids = np.load(validation_data_ids_path)
        training_labels = np.load(training_data_labels_path)
        validation_labels = np.load(validation_data_labels_path)

        training_labels = training_labels.flatten()
        validation_labels = validation_labels.flatten()
        training_labels -= training_labels.min()
        validation_labels -= validation_labels.min()

        training_dataset = TextDataset(training_token_ids, training_labels)
        validation_dataset = TextDataset(validation_token_ids, validation_labels)
        training_data_sampler = SortishSampler(data_source=training_token_ids, key=lambda x: len(training_token_ids[x]),
                                               bs=batch_size // 2)
        validation_data_sampler = SortSampler(data_source=validation_token_ids,
                                              key=lambda x: len(validation_token_ids[x]))
        training_dataloader = DataLoader(dataset=training_dataset, batch_size=batch_size // 2, transpose=True,
                                         num_workers=1, pad_idx=1, sampler=training_data_sampler)
        validation_dataloader = DataLoader(dataset=validation_dataset, batch_size=batch_size, transpose=True,
                                           num_workers=1, pad_idx=1, sampler=validation_data_sampler)
        self.model_data = ModelData(path=classifier_data_dir, trn_dl=training_dataloader, val_dl=validation_dataloader)

    def initialize_learner(self):
        optimization_function = partial(optim.Adam, betas=(0.8, 0.99))

        self.learner = RNN_Learner(data=self.model_data, models=TextModel(to_gpu(self.model)),
                                   opt_fn=optimization_function)
        self.learner.reg_fn = partial(seq2seq_reg, alpha=2, beta=1)
        self.learner.clip = 25.
        self.learner.metrics = [accuracy]

    def load_finetuned_language_model_weights(self, finetuned_language_model_encoder_path):
        load_model(self.learner.model[0], finetuned_language_model_encoder_path)

    def train(self, trained_classifier_save_path, learning_rates=np.array([1e-4, 1e-4, 1e-4, 1e-3, 1e-2]),
              weight_decay=1e-6, cycle_length=14):
        self.learner.freeze_to(-1)
        self.learner.fit(learning_rates, n_cycle=1, wds=weight_decay, cycle_len=1, use_clr=(8, 3))
        self.learner.freeze_to(-2)
        self.learner.fit(learning_rates, n_cycle=1, wds=weight_decay, cycle_len=1, use_clr=(8, 3))

        self.learner.unfreeze()
        self.learner.fit(learning_rates, n_cycle=1, wds=weight_decay, cycle_len=cycle_length, use_clr=(32, 10))
        save_model(self.learner.model, trained_classifier_save_path)

    def load_trained_classifier_weights(self, trained_classifier_path):
        self.model.load_state_dict(torch.load(trained_classifier_path, map_location=lambda storage, loc: storage))

    def predict(self, text):
        self.model.reset()
        self.model.eval()

        input_string = 'xbos xfld 1 ' + text
        texts = [input_string]
        tokens = self.tokenizer.proc_all(texts)
        encoded_tokens = [self.inspire_data_stoi[p] for p in tokens[0]]
        token_array = np.reshape(np.array(encoded_tokens), (-1, 1))
        token_array = Variable(torch.from_numpy(token_array))
        prediction_scores = self.model(token_array)
        prediction_scores_numpy = prediction_scores[0].data.cpu().numpy()

        return numpy_softmax(prediction_scores_numpy[0])[0]


def numpy_softmax(x):
    if x.ndim == 1:
        x = x.reshape((1, -1))
    max_x = np.max(x, axis=1).reshape((-1, 1))
    exp_x = np.exp(x - max_x)
    return exp_x / np.sum(exp_x, axis=1).reshape((-1, 1))



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.
#
# Modified from the fastai library (https://github.com/fastai/fastai).

"""Classifier API."""


from flask import current_app
from inspire_classifier.domain.models import (
    Classifier,
    LanguageModel
)
from inspire_classifier.domain.preprocessor import (
    generate_and_save_classifier_tokens,
    generate_and_save_language_model_tokens,
    map_and_save_tokens_to_ids_for_classifier,
    map_and_save_tokens_to_ids_for_language_model,
    split_and_save_data_for_language_model_and_classifier
)
from inspire_classifier.utils import path_for
import numpy as np
import requests


def create_directories():
    """Create the project data and model directories"""
    path_for('classifier_data').mkdir(parents=True, exist_ok=True)
    path_for('language_model_data').mkdir(parents=True, exist_ok=True)
    path_for('classifier_model').mkdir(parents=True, exist_ok=True)
    (path_for('language_model') / 'wikitext_103').mkdir(exist_ok=True, parents=True)


def preprocess_and_save_data():
    """
    Prepares the data for training.
    """
    try:
        split_and_save_data_for_language_model_and_classifier(
            dataframe_path=path_for('dataframe'), language_model_data_dir=path_for('language_model_data'),
            classifier_data_dir=path_for('classifier_data'),
            val_fraction=current_app.config['CLASSIFIER_VALIDATION_DATA_FRACTION']
        )
    except IOError as error:
        raise IOError('Training dataframe not found.') from error

    try:
        generate_and_save_language_model_tokens(language_model_data_dir=path_for('language_model_data'))
    except IOError as error:
        raise IOError('Language Model data directory does not exist.') from error

    try:
        map_and_save_tokens_to_ids_for_language_model(
            language_model_data_dir=path_for('language_model_data'), data_itos_path=path_for('data_itos'),
            max_vocab_size=current_app.config['CLASSIFIER_MAXIMUM_VOCABULARY_SIZE'],
            minimum_frequency=current_app.config['CLASSIFIER_MINIMUM_WORD_FREQUENCY']
        )
    except IOError as error:
        raise IOError('Language Model data directory or the data directory do not exist.') from error

    try:
        generate_and_save_classifier_tokens(classifier_data_dir=path_for('classifier_data'))
    except IOError as error:
        raise IOError('Classifier data directory does not exist.') from error

    try:
        map_and_save_tokens_to_ids_for_classifier(classifier_data_dir=path_for('classifier_data'),
                                                  data_itos_path=path_for('data_itos'))
    except IOError as error:
        raise IOError('Classifier data directory or the data ITOS does not exist.') from error


def finetune_and_save_language_model():
    """
    Finetunes the pretrained (on wikitext103) language model on our dataset.
    """
    try:
        language_model = LanguageModel(
            training_data_ids_path=path_for('language_model_data') / 'training_token_ids.npy',
            validation_data_ids_path=path_for('language_model_data') / 'validation_token_ids.npy',
            language_model_model_dir=path_for('language_model_data'),
            data_itos_path=path_for('data_itos'), cuda_device_id=current_app.config['CLASSIFIER_CUDA_DEVICE_ID'],
            batch_size=current_app.config['CLASSIFIER_LANGUAGE_MODEL_BATCH_SIZE']
        )
    except IOError as error:
        raise IOError('Training files, language model data directory, or data ITOS do not exist.') from error

    if not path_for('pretrained_language_model').exists():
        wikitext103_language_model_response = requests.get(
            current_app.config['CLASSIFIER_WIKITEXT103_LANGUAGE_MODEL_URL'], allow_redirects=True)
        wikitext103_language_model_response.raise_for_status()
        with open(path_for('pretrained_language_model'), 'wb') as fd:
            fd.write(wikitext103_language_model_response.content)
    if not path_for('wikitext103_itos').exists():
        wikitext103_itos_response = requests.get(current_app.config['CLASSIFIER_WIKITEXT103_ITOS_URL'],
                                                 allow_redirects=True)
        wikitext103_itos_response.raise_for_status()
        with open(path_for('wikitext103_itos'), 'wb') as fd:
            fd.write(wikitext103_itos_response.content)

    try:
        language_model.load_pretrained_language_model_weights(
            pretrained_language_model_path=path_for('pretrained_language_model'),
            wikitext103_itos_path=path_for('wikitext103_itos')
        )
    except IOError as error:
        raise IOError('Wikitext103 pretrained language model and Wikitext103 ITOS do not exist.') from error

    try:
        language_model.train(finetuned_language_model_encoder_save_path=path_for('finetuned_language_model_encoder'),
                             cycle_length=current_app.config['CLASSIFIER_LANGUAGE_MODEL_CYCLE_LENGTH'])
    except IOError as error:
        raise IOError('Unable to save the finetuned language model.') from error


def train_and_save_classifier():
    """
    Trains the classifier on our dataset and save the weights.
    """
    try:
        classifier = Classifier(data_itos_path=path_for('data_itos'),
                                number_of_classes=3, cuda_device_id=current_app.config['CLASSIFIER_CUDA_DEVICE_ID'])
    except IOError as error:
        raise IOError('Data ITOS not found.') from error

    try:
        classifier.load_training_and_validation_data(
            training_data_ids_path=path_for('classifier_data') / 'training_token_ids.npy',
            training_data_labels_path=path_for('classifier_data') / 'training_labels.npy',
            validation_data_ids_path=path_for('classifier_data') / 'validation_token_ids.npy',
            validation_data_labels_path=path_for('classifier_data') / 'validation_labels.npy',
            classifier_data_dir=path_for('classifier_data'),
            batch_size=current_app.config['CLASSIFIER_CLASSIFIER_BATCH_SIZE']
        )
    except IOError as error:
        raise IOError('Training and Validation data for Classifier not found.') from error

    classifier.initialize_learner()

    try:
        classifier.load_finetuned_language_model_weights(
            finetuned_language_model_encoder_path=path_for('finetuned_language_model_encoder')
        )
    except IOError as error:
        raise IOError('Finetuned Language Model Encoder does not exist.') from error

    try:
        classifier.train(trained_classifier_save_path=path_for('trained_classifier'),
                         cycle_length=current_app.config['CLASSIFIER_CLASSIFIER_CYCLE_LENGTH'])
    except IOError as error:
        raise IOError('Unable to save the trained classifier.') from error


def train():
    """
    Runs the complete training pipeline.
    """
    create_directories()
    preprocess_and_save_data()
    finetune_and_save_language_model()
    train_and_save_classifier()


def predict_coreness(title, abstract):
    """
    Predicts class-wise probabilities given the title and abstract.
    """
    text = title + ' ' + abstract
    categories = ['rejected', 'non_core', 'core']
    try:
        classifier = Classifier(data_itos_path=path_for('data_itos'),
                                number_of_classes=3, cuda_device_id=current_app.config['CLASSIFIER_CUDA_DEVICE_ID'])
    except IOError as error:
        raise IOError('Data ITOS not found.') from error

    try:
        classifier.load_trained_classifier_weights(path_for('trained_classifier'))
    except IOError as error:
        raise IOError('Could not load the trained classifier weights.') from error

    class_probabilities = classifier.predict(text)
    assert len(class_probabilities) == 3

    predicted_class = categories[np.argmax(class_probabilities)]
    output_dict = {'prediction': predicted_class}
    output_dict['scores'] = dict(zip(categories, class_probabilities))

    return output_dict



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2014-2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Workflow for processing single arXiv records harvested."""

from __future__ import absolute_import, division, print_function

from workflow.patterns.controlflow import (
    IF,
    IF_NOT,
    IF_ELSE,
)

from inspirehep.modules.workflows.tasks.refextract import extract_journal_info
from inspirehep.modules.workflows.tasks.arxiv import (
    arxiv_author_list,
    arxiv_package_download,
    arxiv_derive_inspire_categories,
    populate_arxiv_document,
)
from inspirehep.modules.workflows.tasks.actions import (
    count_reference_coreness,
    download_documents,
    is_arxiv_paper,
    is_submission,
    mark,
    normalize_journal_titles,
    populate_journal_coverage,
    populate_submission_document,
    refextract,
    save_workflow,
    validate_record,
)
from inspirehep.modules.workflows.tasks.upload import set_schema

from inspirehep.modules.workflows.tasks.makejson import makejson


ENHANCE_RECORD = [
    IF(
        is_arxiv_paper,
        [
            populate_arxiv_document,
            arxiv_package_download,
            arxiv_derive_inspire_categories,
            arxiv_author_list("authorlist2marcxml.xsl"),
        ]
    ),
    IF(
        is_submission,
        populate_submission_document,
    ),
    download_documents,
    normalize_journal_titles,
    refextract,
    count_reference_coreness,
    extract_journal_info,
    populate_journal_coverage,
]

INIT_MARKS = [
    mark('auto-approved', None),
    mark('already-in-holding-pen', None),
    mark('previously_rejected', None),
    mark('is-update', None),
    mark('stopped-matched-holdingpen-wf', None),
    mark('approved', None),
    mark('unexpected-workflow-path', None),
    save_workflow
]

PRE_PROCESSING = [
    # Make sure schema is set for proper indexing in Holding Pen
    set_schema,
    INIT_MARKS,
    validate_record('hep')
]

MAKE_JSON = [
	makejson
]


class Article(object):
    """Article ingestion workflow for Literature collection."""
    name = "HEP"
    data_type = "hep"

    workflow = (
        PRE_PROCESSING +
        ENHANCE_RECORD +
        MAKE_JSON
    )
